//this example is concerned with training a simple RBM and the varying degrees of freedom one
//has to get good performance. For time constraints, we only use a toy problem here.


//used for training the RBM
#include <shark/Unsupervised/RBM/BinaryRBM.h>
#include <shark/ObjectiveFunctions/Regularizer.h>//two-norm regularization ("weight decay")
#include <shark/Algorithms/GradientDescent/SteepestDescent.h>

//the problem
#include <shark/Unsupervised/RBM/Problems/BarsAndStripes.h>

//for evaluation
#include <shark/Unsupervised/RBM/analytics.h>
#include <iostream>

using namespace shark;
using namespace std;

int main(){

	//we first create the problem. in this tutorial, we use BarsAndStripes
	BarsAndStripes problem;
	UnlabeledData<RealVector> data = problem.data();
	
	//some constants needed for training
	size_t numberOfHidden = 16;//hidden units of the rbm
	size_t numberOfVisible = problem.inputDimension();//visible units of the rbm - same as number of inputs

	//create rbm with simple binary units using the global random number generator
	BinaryRBM rbm(Rng::globalRng);
	Rng::seed(42);//initialize random number generator
	rbm.setStructure(numberOfVisible,numberOfHidden);
	
	//create derivative estimator to optimize the rbm
	//we want a simple vanilla CD-1.
	BinaryCD estimator(&rbm);
	TwoNormRegularizer regularizer;
	//0.0 is the regularization strength. 0.0 means no regularization. choose >= 0.0
	estimator.setRegularizer(0.0,&regularizer);
	estimator.setK(1);//number of sampling steps
	estimator.setData(data);//the data used for optimization

	//create and configure optimizer
	SteepestDescent optimizer;
	optimizer.setMomentum(0.0);//momentum term which does a bit of smoothing
	optimizer.setLearningRate(0.1);//learning rate of the algorithm
	
	//now we train the rbm and evaluate the mean negative log-likelihood at the end
	unsigned int numIterations = 20000;//iterations for training
	unsigned int numTrials = 1;//number of trials for training, we print out the mean at the end
	double meanResult = 0;//the mean of the likelihood of the last iteration over all trials
	for(unsigned int trial = 0; trial != numTrials; ++trial) {
		initRandomUniform(rbm,-0.1,0.1);//initialize weights uniformly in [-0.1,0.1]
		optimizer.init(estimator);

		for(unsigned int iteration = 0; iteration != numIterations; ++iteration) {
			optimizer.step(estimator);
			//print the current likelihood at every hundreth iteration
			//this is only possible for small problems!
			if(iteration % (numIterations/100) == 0){
				double likelihood = negativeLogLikelihood(rbm,data);
				std::cout<<trial<<" "<<iteration<<" "<<likelihood<<std::endl;
			}
		}
		//print out result of last iteration
		double likelihood = negativeLogLikelihood(rbm,data);
		std::cout<<trial<<" "<<likelihood<<std::endl;
		meanResult +=likelihood;
	}
	meanResult /= numTrials;

	//print the mean performance
	cout << "RESULTS: " << std::endl;
	cout << "======== " << std::endl;
	cout << "mean negative log likelihood: " << meanResult << std::endl;

}

