//this example is concerned with training a simple RBM and the varying degrees of freedom one
//has to get good performance. For time constraints, we only use a toy problem here.


//used for training the RBM
#include <shark/Unsupervised/RBM/BinaryRBM.h>
#include <shark/ObjectiveFunctions/Regularizer.h>//two-norm regularization ("weight decay")
#include <shark/Algorithms/GradientDescent/SteepestDescent.h>

//used for training the neural network
#include <shark/Models/FFNet.h>
#include<shark/ObjectiveFunctions/Loss/CrossEntropy.h> // loss during network training
#include<shark/ObjectiveFunctions/ErrorFunction.h> //error function to connect data model and loss

//the problem
#include "BarsAndStripesClassification.h"

//for evaluation
#include<shark/ObjectiveFunctions/Loss/ZeroOneLoss.h> //loss for classification
#include <shark/Unsupervised/RBM/analytics.h>
#include <iostream>

using namespace shark;
using namespace std;


//trains first an RBM and transforms it into a neural network.
FFNet<LogisticNeuron,LinearNeuron> unsupervisedPretraining(
	UnlabeledData<RealVector> const& data,
	std::size_t numberOfHidden
){
	//the first part of this is basically the same as the other tutorial.
	
	//create rbm with simple binary units using the global random number generator
	BinaryRBM rbm(Rng::globalRng);
	rbm.setStructure(dataDimension(data),numberOfHidden);
	initRandomUniform(rbm,-0.1,0.1);//initialize weights uniformly in [-0.1,0.1]
	
	//create derivative to optimize the rbm
	//we want a simple vanilla CD-1.
	BinaryCD estimator(&rbm);
	TwoNormRegularizer regularizer;
	//0.0 is the regularization strength. 0.0 means no regularization. choose as >= 0.0
	estimator.setRegularizer(0.0,&regularizer);
	estimator.setK(1);//number of sampling steps
	estimator.setData(data);//the data used for optimization

	//create and configure optimizer
	SteepestDescent optimizer;
	optimizer.setMomentum(0.0);//momentum term which does a bit of smoothing
	optimizer.setLearningRate(0.1);//learning rate of the algorithm
	
	//now we train the rbm and evaluate the mean negative log-likelihood at the end
	unsigned int numIterations = 10000;//iterations for training
	optimizer.init(estimator);
	for(unsigned int iteration = 0; iteration != numIterations; ++iteration) {
		optimizer.step(estimator);
	}
	double likelihood = negativeLogLikelihood(rbm,data);
	std::cout<<"unsupervised training done. Likelihood: "<<likelihood<<std::endl;
	
	// now we need to convert this into the lower layer of a FFNet. 
	
	// first create a network with a single hidden layer, a normal topology
	// without shortcuts and bias neurons
	rbm.setParameterVector(optimizer.solution().point);
	FFNet<LogisticNeuron,LinearNeuron> network;
	network.setStructure(dataDimension(data),numberOfHidden,1,FFNetStructures::Normal,true);
	//we don't set all weights, so initialize randomly first
	initRandomUniform(network,-0.1,0.1);
	//set the other weights to the trained ones
	network.setLayer(0,rbm.weightMatrix(),rbm.hiddenNeurons().bias());
	return network;
}

int main(){
	
	//hyper parameters
	std::size_t numberOfHidden = 8;
	unsigned numberOfSupervisedLearningSteps = 10000;
	Rng::seed(43);//initialize random number generator

	//we first create the problem. in this tutorial, we use the classification version of BarsAndStripes
	//we create a test set to test generalization performance
	BarsAndStripesClassification problem;
	LabeledData<RealVector,unsigned int> data = problem.data();
	data.shuffle();
	LabeledData<RealVector,unsigned int> test = splitAtElement(data,static_cast<std::size_t>(0.5*data.numberOfElements()));
	
	
	//do the unsupervised pre-training.
	FFNet<LogisticNeuron,LinearNeuron> network = unsupervisedPretraining(
		data.inputs(),numberOfHidden
	);
	//in case you want to skip the pre-training, uncomment the following and comment above:
	//FFNet<LogisticNeuron,LinearNeuron> network;
	//network.setStructure(inputDimension(data),numberOfHidden,1,FFNetStructures::Normal,true);
	//initRandomUniform(network,-0.1,0.1);
	
	//create error function
	CrossEntropy loss; // surrogate loss for training
	ErrorFunction<RealVector,unsigned int> error(data,&network,&loss);
	TwoNormRegularizer regularizer;
	//0.0 is the regularization strength. 0.0 means no regularization. choose as >= 0.0
	error.setRegularizer(0.0,&regularizer);
	
	// loss for evaluation
	ZeroOneLoss<unsigned int, RealVector> loss01(0.5); // classification error, output is thresholded at 1/2
	
	//initialize the optimizer and optimize the network
	SteepestDescent optimizer;
	optimizer.setLearningRate(0.5);
	optimizer.init(error);
	for(unsigned iteration = 0; iteration != numberOfSupervisedLearningSteps; ++iteration){
		optimizer.step(error);
		if(iteration % (numberOfSupervisedLearningSteps/10) == 0){
			network.setParameterVector(optimizer.solution().point); // set weights to weights found by learning
			Data<RealVector> prediction = network(test.inputs());
			cout << iteration<<" "<<optimizer.solution().value<< " " << loss01.eval(test.labels(), prediction) << endl;
		}
	}

	// evaluate solution found by training
	network.setParameterVector(optimizer.solution().point); // set weights to weights found by learning
	Data<RealVector> prediction = network(test.inputs());
	cout << "classification error after learning:\t" << loss01.eval(test.labels(), prediction) << endl;

}

