//this example is concerned with training a simple RBM and the varying degrees of freedom one
//has to get good performance. For time constraints, we only use a toy problem here.


//used for training the RBM
#include <shark/Unsupervised/RBM/BinaryRBM.h>
#include <shark/ObjectiveFunctions/Regularizer.h>//two-norm regularization ("weight decay")
#include <shark/Algorithms/GradientDescent/SteepestDescent.h>


//used for classifying the features
//we will use a linear SVM here.
#include <shark/Models/Kernels/LinearKernel.h>
#include <shark/Algorithms/Trainers/CSvmTrainer.h>
#include <shark/Algorithms/Trainers/LDA.h>
//the problem
#include "BarsAndStripesClassification.h"

//for evaluation
#include<shark/ObjectiveFunctions/Loss/ZeroOneLoss.h> //loss for classification
#include <shark/Unsupervised/RBM/analytics.h>
#include <iostream>

using namespace shark;
using namespace std;


//trains first an RBM and transforms the data onto the features learned by the hidden units
BinaryRBM unsupervisedTraining(
	UnlabeledData<RealVector> const& data,
	std::size_t numberOfHidden
){
	//the first part of this is basically the same as the other tutorial.
	
	//create rbm with simple binary units using the global random number generator
	BinaryRBM rbm(Rng::globalRng);
	rbm.setStructure(dataDimension(data),numberOfHidden);
	initRandomUniform(rbm,-0.1,0.1);//initialize weights uniformly in [-0.1,0.1]
	
	//create derivative to optimize the rbm
	//we want a simple vanilla CD-1 with weight-decay.
	BinaryCD estimator(&rbm);
	TwoNormRegularizer regularizer;
	//0.0 is the regularization strength. 0.0 means no regularization. choose as >= 0.0
	estimator.setRegularizer(0.0,&regularizer);
	estimator.setK(1);//number of sampling steps
	estimator.setData(data);//the data used for optimization

	//create and configure optimizer
	SteepestDescent optimizer;
	optimizer.setMomentum(0.0);//momentum term which does a bit of smoothing
	optimizer.setLearningRate(0.1);//learning rate of the algorithm
	
	//now we train the rbm and evaluate the mean negative log-likelihood at the end
	unsigned int numIterations = 10000;//iterations for training
	optimizer.init(estimator);
	for(unsigned int iteration = 0; iteration != numIterations; ++iteration) {
		optimizer.step(estimator);
	}
	double likelihood = negativeLogLikelihood(rbm,data);
	std::cout<<"unsupervised training done. Likelihood: "<<likelihood<<std::endl;
	
	// now we create the feature map
	rbm.setParameterVector(optimizer.solution().point);
	return rbm;
}

int main(){
	
	//hyper parameters
	std::size_t numberOfHidden = 8;
	unsigned numberOfSupervisedLearningSteps = 10000;
	Rng::seed(43);//initialize random number generator

	//we first create the problem. in this tutorial, we use the classification version of BarsAndStripes
	//we create a test set to test generalization performance
	BarsAndStripesClassification problem;
	LabeledData<RealVector,unsigned int> data = problem.data();
	data.shuffle();
	LabeledData<RealVector,unsigned int> test = splitAtElement(data,static_cast<std::size_t>(0.5*data.numberOfElements()));
	
	
	//do the unsupervised pre-training.
	BinaryRBM rbm = unsupervisedTraining(
		data.inputs(),numberOfHidden
	);
	//transform the datasets by projecting them on the learned features
	rbm.evaluationType(true,true);//we compute the direction visible->hidden and want the features.
	LabeledData<RealVector,unsigned int> dataProj = transformInputs(data,rbm);
	LabeledData<RealVector,unsigned int> testProj = transformInputs(test,rbm);
	
	//we train on the learned features and without to compare the results
	LDA trainer;
	trainer.setRegularization(0.0);
	LinearClassifier<RealVector> model;
	LinearClassifier<RealVector> modelBaseline;
	trainer.train(model,dataProj);
	trainer.train(modelBaseline,data);

	// evaluate models
	ZeroOneLoss<> loss01;
	//training set
	Data<unsigned int> predictionTrain = model(dataProj.inputs());
	Data<unsigned int> predictionBaselineTrain = modelBaseline(data.inputs());
	cout << "classification error,features: " << loss01.eval(dataProj.labels(), predictionTrain) << endl;
	cout << "classification error,baseline: " << loss01.eval(data.labels(), predictionBaselineTrain) << endl;
	
	Data<unsigned int> prediction = model(testProj.inputs());
	Data<unsigned int> predictionBaseline = modelBaseline(test.inputs());
	cout << "classification error,features: " << loss01.eval(testProj.labels(), prediction) << endl;
	cout << "classification error,baseline: " << loss01.eval(test.labels(), predictionBaseline) << endl;

}

